@@ -1,15 +1,15 @@
 # Authors: Stephane Gaiffas <stephane.gaiffas@gmail.com>
 # License: BSD 3 clause
-
+import os
 import numpy as np
 from numba import jitclass, njit
 from numba import types, _helperlib
-from numba.types import float32, boolean, uint32, string, void, int32
-
+from .types import float32, boolean, uint32, string, void, get_array_2d_type
 from .checks import check_X_y, check_array
-from .sample import SamplesCollection
+from .sample import SamplesCollection, add_samples
 from .tree import TreeClassifier
 from .tree_methods import tree_partial_fit, tree_predict
+from .utils import get_type
 
 spec = [
     ("n_classes", uint32),
@@ -21,9 +21,10 @@ spec = [
     ("dirichlet", float32),
     ("split_pure", boolean),
     ("n_jobs", uint32),
+    ("reserve_samples", uint32),
     ("verbose", boolean),
-    ("trees", types.List(TreeClassifier.class_type.instance_type, reflected=True)),
-    ("samples", SamplesCollection.class_type.instance_type),
+    ("trees", types.List(get_type(TreeClassifier), reflected=True)),
+    ("samples", get_type(SamplesCollection)),
     ("iteration", uint32),
 ]
 
@@ -44,6 +45,7 @@ class AMFClassifierNoPython(object):
         dirichlet,
         split_pure,
         n_jobs,
+        reserve_samples,
         verbose,
     ):
         self.n_classes = n_classes
@@ -55,10 +57,11 @@ class AMFClassifierNoPython(object):
         self.dirichlet = dirichlet
         self.split_pure = split_pure
         self.n_jobs = n_jobs
+        self.reserve_samples = reserve_samples
         self.verbose = verbose
         self.iteration = 0
 
-        samples = SamplesCollection()
+        samples = SamplesCollection(self.reserve_samples)
         self.samples = samples
 
         # TODO: reflected lists will be replaced by typed list soon...
@@ -78,55 +81,51 @@ class AMFClassifierNoPython(object):
         self.trees = trees
 
 
-@njit(
-    void(AMFClassifierNoPython.class_type.instance_type, float32[:, ::1], float32[::1])
-)
+@njit(void(get_type(AMFClassifierNoPython), get_array_2d_type(float32), float32[::1],))
 def partial_fit(forest, X, y):
     n_samples_batch, n_features = X.shape
-    # We need at least the actual number of nodes + twice the extra samples
-
-    # Let's reserve nodes in advance
-    for tree in forest.trees:
-        n_nodes_reserved = tree.nodes.n_nodes + 2 * n_samples_batch
-        tree.nodes.reserve_nodes(n_nodes_reserved)
-
     # First, we save the new batch of data
     n_samples_before = forest.samples.n_samples
-    forest.samples.append(X, y)
-
+    # Add the samples in the forest
+    add_samples(forest.samples, X, y)
     for i in range(n_samples_before, n_samples_before + n_samples_batch):
         # Then we fit all the trees using all new samples
         for tree in forest.trees:
             tree_partial_fit(tree, i)
         forest.iteration += 1
 
-    # def predict(self, X, scores):
-    #     scores.fill(0.0)
-    #     n_samples_batch, _ = X.shape
-    #     if self.iteration > 0:
-    #         scores_tree = np.empty(self.n_classes, float32)
-    #         for i in range(n_samples_batch):
-    #             # print('i:', i)
-    #             scores_i = scores[i]
-    #             x_i = X[i]
-    #             # print('x_i:', x_i)
-    #             # The prediction is simply the average of the predictions
-    #             for tree in self.trees:
-    #                 tree_predict(tree, x_i, scores_tree, self.use_aggregation)
-    #                 # print('scores_tree:', scores_tree)
-    #                 scores_i += scores_tree
-    #             scores_i /= self.n_estimators
-    #             # print('scores_i:', scores_i)
-    #     else:
-    #         raise RuntimeError("You must call ``partial_fit`` before ``predict``.")
+
+# TODO: code predict
+# def predict(self, X, scores):
+#     scores.fill(0.0)
+#     n_samples_batch, _ = X.shape
+#     if self.iteration > 0:
+#         scores_tree = np.empty(self.n_classes, float32)
+#         for i in range(n_samples_batch):
+#             # print('i:', i)
+#             scores_i = scores[i]
+#             x_i = X[i]
+#             # print('x_i:', x_i)
+#             # The prediction is simply the average of the predictions
+#             for tree in self.trees:
+#                 tree_predict(tree, x_i, scores_tree, self.use_aggregation)
+#                 # print('scores_tree:', scores_tree)
+#                 scores_i += scores_tree
+#             scores_i /= self.n_estimators
+#             # print('scores_i:', scores_i)
+#     else:
+#         raise RuntimeError("You must call ``partial_fit`` before ``predict``.")
 
 
 @njit(
     void(
-        AMFClassifierNoPython.class_type.instance_type, float32[:, ::1], float32[:, ::1]
+        get_type(AMFClassifierNoPython),
+        get_array_2d_type(float32),
+        get_array_2d_type(float32),
     )
 )
 def predict_proba(forest, X, scores):
+    # TODO: use predict_proba_tree from below ? Or put it in the tree ?
     scores.fill(0.0)
     n_samples_batch, _ = X.shape
 
@@ -145,76 +144,54 @@ def predict_proba(forest, X, scores):
         # print('scores_i:', scores_i)
 
 
+@njit(
+    get_array_2d_type(float32)(
+        get_type(AMFClassifierNoPython), uint32, get_array_2d_type(float32)
+    )
+)
+def predict_proba_tree(forest, idx_tree, X):
+    n_samples_batch, _ = X.shape
+    scores = np.empty((n_samples_batch, forest.n_classes), dtype=float32)
+    tree = forest.trees[idx_tree]
+    for i in range(n_samples_batch):
+        scores_i = scores[i]
+        x_i = X[i]
+        tree_predict(tree, x_i, scores_i, forest.use_aggregation)
+    return scores
+
+
 class AMFClassifier(object):
     """Aggregated Mondrian Forest classifier for online learning. This algorithm
-    is truly online, in the sense that a single pass is performed, and one can
-    ask for predictions anytime.
+    is truly online, in the sense that a single pass is performed, and that predictions
+    can be produced anytime.
 
-    Parameters
-    ----------
-    n_classes : `int`
-        Number of excepted classes in the labels. This is required since we
-        don't know the number of classes in advance in a online setting.
+    Each node in a tree predicts according to the distribution of the labels
+    it contains. This distribution is regularized using a "Jeffreys" prior
+    with parameter ``dirichlet``. For each class with `count` labels in the
+    node and `n_samples` samples in it, the prediction of a node is given by
 
-    n_estimators : `int`, optional (default=10)
-        Number of trees to grow in the forest.
+        (count + dirichlet) / (n_samples + dirichlet * n_classes)
 
-    step : `float`, optional (default=1)
-        Step-size for the aggregation weights. Default is 1 for classification,
-        which is typically the best choice.
+    The prediction for a sample is computed as the aggregated predictions of all the
+    subtrees along the path leading to the leaf node containing the sample. The
+    aggregation weights are exponential weights with learning rate ``step`` and loss
+    ``loss`` when ``use_aggregation`` is ``True``.
 
-    loss : 'str', optional (default='log')
-        The loss used for the computation of the aggregation weights. Only `log`
-        is supported for now, namely the log-loss for multi-class
-        classification.
+    This computation is performed exactly thanks to a context tree weighting algorithm.
+    More details can be found in the paper cited in references below.
 
-    use_aggregation : `bool`, optional (default=True)
-        Whether to use aggregation in each tree. It is highly recommended to
-        leave it as `True`.
+    The final predictions are the average class probabilities predicted by each of the
+    ``n_estimators`` trees in the forest.
 
-    dirichlet : `float` or `None`, optional (default=None)
-        Each node in a tree predicts according to the distribution of the labels
-        it contains. This distribution is regularized using a "Jeffreys" prior
-        with parameter `dirichlet`. For each class with `count` labels in the
-        node and `n_samples` samples in it, the prediction of a node is given by
-        (count + dirichlet) / (n_samples + dirichlet * n_classes).
-
-        Default is dirichlet=0.5 for n_classes=2 and dirichlet=0.01 otherwise.
-
-    split_pure : `bool`, optional (default=False)
-        Whether we split nodes that contain only one class of labels. Default is
-        False.
-
-    n_jobs : `int`, optional (default=1)
-        The number of threads used to grow the tree in parallel. This has no
-        effect for now, the default is n_jobs=1, namely single-threaded.
-
-    random_state : `int` or `None`, optional (default=None)
-        Controls the randomness involved in the trees growing, which is a highly
-        randomized process.
-
-    verbose : `bool`, optional (default=True)
-        Whether a bar showing progress for `partial_fit`, `predict_proba` and
-        `predict` should be used.
-
-    Attributes
-    ----------
-    n_classes = `int`
-        Number of excepted classes in the labels.
-
-    n_features : `int`
-        The number of features from the training dataset (passed to ``fit``)
-
-    # TODO: add missing attributes
-
-    Notes
-    -----
-    Note that all the parameters of AMFClassifier become read-only after the
-    first call to `partial_fit`
+    Note
+    ----
+    All the parameters of ``AMFClassifier`` become **read-only** after the first call
+    to ``partial_fit``
 
     References
     ----------
-    TODO : add the reference of the paper
+    J. Mourtada, S. Gaiffas and E. Scornet, *AMF: Aggregated Mondrian Forests for Online Learning*, arXiv:1906.10529, 2019
+
     """
 
     def __init__(
@@ -227,14 +204,64 @@ class AMFClassifier(object):
         dirichlet=None,
         split_pure=False,
         n_jobs=1,
+        reserve_samples=1024,
         random_state=None,
-        verbose: bool = True,
+        verbose=False,
     ):
+        """Instantiates a `AMFClassifier` instance.
 
+        Parameters
+        ----------
+        n_classes : :obj:`int`
+            Number of expected classes in the labels. This is required since we
+            don't know the number of classes in advance in a online setting.
+
+        n_estimators : :obj:`int`, default = 10
+            The number of trees in the forest.
+
+        step : :obj:`float`, default = 1
+            Step-size for the aggregation weights. Default is 1 for classification with
+            the log-loss, which is usually the best choice.
+
+        loss : {"log"}, default = "log"
+            The loss used for the computation of the aggregation weights. Only "log"
+            is supported for now, namely the log-loss for multi-class
+            classification.
+
+        use_aggregation : :obj:`bool`, default = `True`
+            Controls if aggregation is used in the trees. It is highly recommended to
+            leave it as `True`.
+
+        dirichlet : :obj:`float` or :obj:`None`, default = `None`
+            Regularization level of the class frequencies used for predictions in each
+            node. Default is dirichlet=0.5 for n_classes=2 and dirichlet=0.01 otherwise.
+
+        split_pure : :obj:`bool`, default = `False`
+            Controls if nodes that contains only sample of the same class should be
+            split ("pure" nodes). Default is `False`, namely pure nodes are not split,
+            but `True` can be sometimes better.
+
+        n_jobs : :obj:`int`, default = 1
+            Sets the number of threads used to grow the tree in parallel. The default is
+            n_jobs=1, namely single-threaded. Fow now, this parameter has no effect and
+            only a single thread can be used.
+
+        reserve_samples : :obj:`int`, default = 1024
+            Sets the amount of memory which is pre-allocated each time extra memory is
+            required for new samples and new nodes. Decreasing it can slow down
+            training. If you know that each ``partial_fit`` will be called with
+            approximately `n` samples, you can set reserve_samples = `n` if `n` is
+            large (> 1024).
+
+        random_state : :obj:`int` or :obj:`None`, default = `None`
+            Controls the randomness involved in the trees.
+
+        verbose : :obj:`bool`, default = `False`
+            Controls the verbosity when fitting and predicting.
+        """
         # We will instantiate the numba class when data is passed to
-        # `partial_fit`, since we need to know about `n_features` among others
+        # `partial_fit`, since we need to know about `n_features` among others things
         self.no_python = None
-
         self.n_classes = n_classes
         self._n_features = None
         self.n_estimators = n_estimators
@@ -252,15 +279,36 @@ class AMFClassifier(object):
 
         self.split_pure = split_pure
         self.n_jobs = n_jobs
-
-        # TODO: deal with random_state
+        self.reserve_samples = reserve_samples
         self.random_state = random_state
         self.verbose = verbose
-
         self._classes = set(range(n_classes))
 
-    def partial_fit(self, X, y):
-        # TODO: write the docstring
+        if os.getenv("NUMBA_DISABLE_JIT", None) == "1":
+            self._using_numba = False
+        else:
+            self._using_numba = True
+
+    def partial_fit(self, X, y, classes=None):
+        """Updates the classifier with the given batch of samples.
+
+        Parameters
+        ----------
+        X : :obj:`np.ndarray`, shape=(n_samples, n_features)
+            Input features matrix.
+
+        y : :obj:`np.ndarray`
+            Input labels vector.
+
+        classes : :obj:`None`
+            Must not be used, only here for backwards compatibility
+
+        Returns
+        -------
+        output : :obj:`AMFClassifier`
+            Updated instance of :obj:`AMFClassifier`
+
+        """
         # First,ensure that X and y are C-contiguous and with float32 dtype
         X, y = check_X_y(
             X,
@@ -301,6 +349,7 @@ class AMFClassifier(object):
                 self.dirichlet,
                 self.split_pure,
                 self.n_jobs,
+                self.reserve_samples,
                 self.verbose,
             )
         else:
@@ -316,17 +365,18 @@ class AMFClassifier(object):
         return self
 
     def predict_proba(self, X):
-        """Predict class for given samples
+        """Predicts the class probabilities for the given features vectors
 
         Parameters
         ----------
-        X : `np.ndarray`, shape=(n_samples, n_features)
-            Features matrix to predict for
+        X : :obj:`np.ndarray`, shape=(n_samples, n_features)
+            Input features matrix to predict for.
 
         Returns
         -------
-        output : `np.ndarray`, shape=(n_samples, n_classes)
-            Returns the predicted class probabilities
+        output : :obj:`np.ndarray`, shape=(n_samples, n_classes)
+            Returns the predicted class probabilities for the input features
+
         """
         import numpy as np
 
@@ -344,9 +394,8 @@ class AMFClassifier(object):
             ensure_min_features=1,
             estimator="AMFClassifier",
         )
-
         n_samples, n_features = X.shape
-        scores = np.empty((n_samples, self.n_classes), dtype="float32")
+        scores = np.zeros((n_samples, self.n_classes), dtype="float32")
         if not self.no_python:
             raise RuntimeError(
                 "You must call `partial_fit` before calling `predict_proba`"
@@ -362,6 +411,61 @@ class AMFClassifier(object):
         self._put_back_random_state()
         return scores
 
+    def predict_proba_tree(self, X, tree):
+        """Predicts the class probabilities for the given features vectors using a
+        single tree at given index ``tree``. Should be used only for debugging or
+        visualisation purposes.
+        
+        Parameters
+        ----------
+        X : :obj:`np.ndarray`, shape=(n_samples, n_features)
+            Input features matrix to predict for.
+
+        tree : :obj:`int`
+            Index of the tree, must be between 0 and ``n_estimators`` - 1
+
+        Returns
+        -------
+        output : :obj:`np.ndarray`, shape=(n_samples, n_classes)
+            Returns the predicted class probabilities for the input features
+
+        """
+        # TODO: unittests for this method
+        if not self.no_python:
+            raise RuntimeError(
+                "You must call `partial_fit` before calling `predict_proba`"
+            )
+        else:
+            X = check_array(
+                X,
+                accept_sparse=False,
+                accept_large_sparse=False,
+                dtype=["float32"],
+                order="C",
+                copy=False,
+                force_all_finite=True,
+                ensure_2d=True,
+                allow_nd=False,
+                ensure_min_samples=1,
+                ensure_min_features=1,
+                estimator="AMFClassifier",
+            )
+            n_samples, n_features = X.shape
+            if n_features != self.n_features:
+                raise ValueError(
+                    "`partial_fit` was called with n_features=%d while `predict_proba` "
+                    "received n_features=%d" % (self.n_features, n_features)
+                )
+            if not isinstance(tree, int):
+                raise ValueError("`tree` must be of integer type")
+            if tree < 0 or tree >= self.n_estimators:
+                raise ValueError("`tree` must be between 0 and `n_estimators` - 1")
+
+            self._set_random_state()
+            scores = predict_proba_tree(self.no_python, tree, X)
+            self._put_back_random_state()
+            return scores
+
     # def predict(self, X):
     #     if not self._fitted:
     #         raise RuntimeError("You must call ``fit`` before")
@@ -374,21 +478,25 @@ class AMFClassifier(object):
         # This uses a trick by Alexandre Gramfort,
         #   see https://github.com/numba/numba/issues/3249
         if self._random_state >= 0:
-            r = np.random.RandomState(self._random_state)
-            ptr = _helperlib.rnd_get_np_state_ptr()
-            ints, index = r.get_state()[1:3]
-            _helperlib.rnd_set_state(ptr, (index, [int(x) for x in ints]))
-            self._ptr = ptr
-            self._r = r
+            if self._using_numba:
+                r = np.random.RandomState(self._random_state)
+                ptr = _helperlib.rnd_get_np_state_ptr()
+                ints, index = r.get_state()[1:3]
+                _helperlib.rnd_set_state(ptr, (index, [int(x) for x in ints]))
+                self._ptr = ptr
+                self._r = r
+            else:
+                np.random.seed(self._random_state)
 
     def _put_back_random_state(self):
         # This uses a trick by Alexandre Gramfort,
         #   see https://github.com/numba/numba/issues/3249
         if self._random_state >= 0:
-            ptr = self._ptr
-            r = self._r
-            index, ints = _helperlib.rnd_get_state(ptr)
-            r.set_state(("MT19937", ints, index, 0, 0.0))
+            if self._using_numba:
+                ptr = self._ptr
+                r = self._r
+                index, ints = _helperlib.rnd_get_state(ptr)
+                r.set_state(("MT19937", ints, index, 0, 0.0))
 
     def get_nodes_df(self, idx_tree):
         import pandas as pd
@@ -396,7 +504,6 @@ class AMFClassifier(object):
         tree = self.no_python.trees[idx_tree]
         nodes = tree.nodes
         n_nodes = nodes.n_nodes
-
         index = nodes.index[:n_nodes]
         parent = nodes.parent[:n_nodes]
         left = nodes.left[:n_nodes]
@@ -408,8 +515,8 @@ class AMFClassifier(object):
         memory_range_min = nodes.memory_range_min[:n_nodes]
         memory_range_max = nodes.memory_range_max[:n_nodes]
         n_samples = nodes.n_samples[:n_nodes]
-        # weight = nodes.weight[:n_nodes]
-        # log_weight_tree = nodes.log_weight_tree[:n_nodes]
+        weight = nodes.weight[:n_nodes]
+        log_weight_tree = nodes.log_weight_tree[:n_nodes]
         is_leaf = nodes.is_leaf[:n_nodes]
         # is_memorized = nodes.is_memorized[:n_nodes]
         counts = nodes.counts[:n_nodes]
@@ -425,6 +532,8 @@ class AMFClassifier(object):
             "threshold",
             "time",
             "n_samples",
+            "weight",
+            "log_weight_tree",
             "memory_range_min",
             "memory_range_max",
             "counts",
@@ -441,6 +550,8 @@ class AMFClassifier(object):
             "is_leaf": is_leaf,
             "time": time,
             "n_samples": n_samples,
+            "weight": weight,
+            "log_weight_tree": log_weight_tree,
             "memory_range_min": [tuple(t) for t in memory_range_min],
             "memory_range_max": [tuple(t) for t in memory_range_max],
             "counts": [tuple(t) for t in counts],
@@ -450,6 +561,7 @@ class AMFClassifier(object):
 
     @property
     def n_classes(self):
+        """:obj:`int`: Number of expected classes in the labels."""
         return self._n_classes
 
     @n_classes.setter
@@ -468,6 +580,7 @@ class AMFClassifier(object):
 
     @property
     def n_features(self):
+        """:obj:`int`: Number of features used during training."""
         return self._n_features
 
     @n_features.setter
@@ -476,6 +589,7 @@ class AMFClassifier(object):
 
     @property
     def n_estimators(self):
+        """:obj:`int`: Number of trees in the forest."""
         return self._n_estimators
 
     @n_estimators.setter
@@ -494,6 +608,7 @@ class AMFClassifier(object):
 
     @property
     def n_jobs(self):
+        """:obj:`int`: Number of expected classes in the labels."""
         return self._n_jobs
 
     @n_jobs.setter
@@ -508,8 +623,29 @@ class AMFClassifier(object):
             else:
                 self._n_jobs = val
 
+    @property
+    def reserve_samples(self):
+        """:obj:`int`: Amount of memory pre-allocated each time extra memory is
+        required."""
+        return self._reserve_samples
+
+    @reserve_samples.setter
+    def reserve_samples(self, val):
+        if self.no_python:
+            raise ValueError(
+                "You cannot modify `reserve_samples` after calling `partial_fit`"
+            )
+        else:
+            if not isinstance(val, int):
+                raise ValueError("`reserve_samples` must be of type `int`")
+            elif val < 1:
+                raise ValueError("`reserve_samples` must be >= 1")
+            else:
+                self._reserve_samples = val
+
     @property
     def step(self):
+        """:obj:`float`: Step-size for the aggregation weights."""
         return self._step
 
     @step.setter
@@ -526,6 +662,7 @@ class AMFClassifier(object):
 
     @property
     def use_aggregation(self):
+        """:obj:`bool`: Controls if aggregation is used in the trees."""
         return self._use_aggregation
 
     @use_aggregation.setter
@@ -542,6 +679,8 @@ class AMFClassifier(object):
 
     @property
     def dirichlet(self):
+        """:obj:`float` or :obj:`None`: Regularization level of the class
+        frequencies."""
         return self._dirichlet
 
     @dirichlet.setter
@@ -560,6 +699,8 @@ class AMFClassifier(object):
 
     @property
     def split_pure(self):
+        """:obj:`bool`: Controls if nodes that contains only sample of the same class
+        should be split."""
         return self._split_pure
 
     @split_pure.setter
@@ -576,6 +717,7 @@ class AMFClassifier(object):
 
     @property
     def verbose(self):
+        """:obj:`bool`: Controls the verbosity when fitting and predicting."""
         return self._verbose
 
     @verbose.setter
@@ -590,6 +732,7 @@ class AMFClassifier(object):
 
     @property
     def loss(self):
+        """:obj:`str`: The loss used for the computation of the aggregation weights."""
         return "log"
 
     @loss.setter
@@ -598,6 +741,7 @@ class AMFClassifier(object):
 
     @property
     def random_state(self):
+        """:obj:`int` or :obj:`None`: Controls the randomness involved in the trees."""
         if self._random_state == -1:
             return None
         else:
