@@ -3,10 +3,11 @@
 
 import numpy as np
 from numba import jitclass
-from numba.types import float32, boolean, uint32, string
+from .types import float32, boolean, uint32, string
 
-from .node_collection import NodeCollection
+from .node_collection import NodeCollection, add_node
 from .sample import SamplesCollection
+from .utils import get_type
 
 spec_tree_classifier = [
     ("n_classes", uint32),
@@ -16,8 +17,8 @@ spec_tree_classifier = [
     ("use_aggregation", boolean),
     ("dirichlet", float32),
     ("split_pure", boolean),
-    ("samples", SamplesCollection.class_type.instance_type),
-    ("nodes", NodeCollection.class_type.instance_type),
+    ("samples", get_type(SamplesCollection)),
+    ("nodes", get_type(NodeCollection)),
     ("intensities", float32[::1]),
     ("iteration", uint32),
 ]
@@ -48,19 +49,14 @@ class TreeClassifier(object):
         self.samples = samples
 
         # TODO: something more clever than this ?
-        default_reserve = 512
-        self.nodes = NodeCollection(n_features, n_classes, default_reserve)
+        # default_reserve = 512
+        # self.nodes = NodeCollection(n_features, n_classes, default_reserve)
+        reserve_samples = self.samples.reserve_samples
+
+        self.nodes = NodeCollection(n_features, n_classes, reserve_samples)
         self.intensities = np.empty(n_features, dtype=float32)
         self.iteration = 0
         self.add_root()
 
     def add_root(self):
-        self.add_node(0, 0.0)
-
-    def add_node(self, parent, time):
-        return self.nodes.add_node(parent, time)
-
-    def print(self):
-        print("Hello from jitclass")
-        print(self.nodes.n_nodes_reserved)
-        print(self.nodes.n_nodes)
+        add_node(self.nodes, 0, 0.0)
